/*
 * $Header: /home/cvs/jakarta-slide/src/share/org/apache/slide/store/tlock/TLockManager.java,v 1.1.2.1 2004/02/05 16:05:13 mholz Exp $
 * $Revision: 1.1.2.1 $
 * $Date: 2004/02/05 16:05:13 $
 *
 * ====================================================================
 *
 * Copyright 1999-2002 The Apache Software Foundation 
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 *
 */
package org.apache.slide.store.tlock;

import org.apache.slide.store.*;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

/**
 * Handles t-locks (transient locks) for multi-thread synchronization in the stores.
 *
 * @author    michael.hartmeier@softwareag.com
 * @author    peter.nevermann@softwareag.com
 * @version   $Revision: 1.1.2.1 $
 */
public class TLockManager {
    
    private final Map threadMap; // thread-name -> list-of-locks
    private final Map lockMap;   // resource-id -> lock
    private final long timeout;
    
    /**
     * Default constructor.
     *
     * @pre        childStore != null
     * @post       globalCache != null
     * @post       deltas != null
     */
    public TLockManager(long timeout) {
        this.timeout = timeout;
        this.threadMap = new HashMap();
        this.lockMap = new HashMap();
    }
    
    public synchronized void lock(ResourceId resourceId, String uri, int lockType) throws TLockedException {
        TLock lock;
        int oldLockType;
        
        if (lockType != TLock.NO_LOCK) {
            if (uri == null) {
                throw new IllegalArgumentException();
            }
        }
        
        lock = (TLock)lockMap.get(resourceId);
        if (lock != null) {
            oldLockType = lock.acquire(lockType);
            if (oldLockType == TLock.NO_LOCK) {
                if (lockType != TLock.NO_LOCK) {
                    addLock(lock);
//                    System.out.println("@@@ L "+uri+","+lockType);
                }
            }
        } else {
            create(lockType, uri, resourceId);
        }
    }
    
    public synchronized void releaseLocks() {
        List locks = (List)threadMap.get(Thread.currentThread());
        if (locks != null) {
            Iterator i = locks.iterator();
            while( i.hasNext() ) {
                TLock lock = (TLock) i.next();
                lockMap.remove(lock.getResourceId());
                lock.release();
            }
            threadMap.remove(Thread.currentThread());
        }
    }
    
    /**
     * Creates an aquires a new ResourceLock.
     */
    public TLock create(int lockType, String uri, ResourceId resourceId) {
        TLock lock = new TLock(this, timeout, resourceId);
        try {
            if (lock.acquire(lockType) != TLock.NO_LOCK) {
                // we've just created the TLock object - nobody else can see/look ip
                throw new IllegalStateException();
            }
        } catch (TLockedException e) {
            // we've just created the TLock object - nobody else can see/look ip
            throw new IllegalStateException(e.getMessage());
        }
        if (lockType != TLock.NO_LOCK) {
            addLock(lock);
//            System.out.println("@@@ C "+uri+","+lockType);
        }
        return lock;
    }
    
    private void addLock(TLock lock) {
        List locks = (List)threadMap.get(Thread.currentThread());
        if (locks == null) {
            locks = new ArrayList();
            threadMap.put(Thread.currentThread(), locks);
        }
        locks.add(lock);
        lockMap.put(lock.getResourceId(), lock);
    }
}



